import numpy as np
from gwu_nn.activation_functions import SoftmaxActivation
from abc import ABC, abstractmethod


class LossFunction(ABC):
    """Abstract Class for loss functions"""

    def __init__(self):
        super().__init__()

    @abstractmethod
    def loss(cls, y_true, y_pred):
        """Calculates the loss for the given network.

        Args:
            y_true (np.array): the true values
            y_pred (np.array): the predicted values

        Returns:
            np.array(np.array): the network's loss
        """
        pass

    @abstractmethod
    def loss_partial_derivative(cls, y_true, y_pred):
        """Calculates the derivative of the loss for the given network.
        
        Args:
            y_true (np.array): the true values
            y_pred (np.array): the predicted values

        Returns:
            np.array(np.array): the partial derivative for the network's loss
        """
        pass


class MSE(LossFunction):
    """Class for implementing the MSE loss function. Inheirits
    loss and loss_partial_derivative from LossFunction"""
    @classmethod
    def loss(cls, y_true, y_pred):
        """Calculates the MSE for the true vs predicted values
        
        Returns:
            np.array: MSE for each input
        """
        return np.mean(np.power(y_true - y_pred, 2))

    @classmethod
    def loss_partial_derivative(cls, y_true, y_pred):
        """Calculates the derivative of the MSE for each prediction
        
        Returns:
            np.array: Partial derivative of the MSE
        """
        return 2 * (y_pred - y_true) / y_pred.size

class LogLoss(LossFunction):
    """Class for implementing the LogLoss loss function. Inheirits
    loss and loss_partial_derivative from LossFunction
    """
    @classmethod
    def loss(cls, y_true, y_pred):
        """Calculates the LogLoss for the true vs predicted values
        
        Returns:
            np.array: LogLoss for each input
        """
        return np.mean(-np.log(y_pred)*y_true + -np.log(1-y_pred)*(1-y_true))

    @classmethod
    def loss_partial_derivative(cls, y_true, y_pred):
        """Calculates the derivative of the LogLoss for each prediction
        
        Returns:
            np.array: Partial derivative of the LogLoss
        """
        return -np.sum(y_true - y_pred)


class CrossEntropy(LossFunction):
    """Class for implementing the CrossEntropy loss function. Inheirits
    loss and loss_partial_derivative from LossFunction"""
    @classmethod
    def loss(cls, y_true, y_pred):
        """Calculates the CrossEntropy for the true vs predicted classes
        
        Returns:
            np.array: CrossEntropy for each input/class
        """
        return -np.mean(y_true*np.log(y_pred))

    @classmethod
    def loss_partial_derivative(cls, y_true, y_pred):
        """Calculates the derivative of the CrossEntropy for each prediction
        
        Returns:
            np.array: Partial derivative of the CrossEntropy
        """
        m = y_true.shape[0]
        grad = SoftmaxActivation.activation(y_pred)
        grad[range(m), y_true] -= 1
        grad = grad / m
        return grad



from gwu_nn.activation_functions import SigmoidActivation, RELUActivation, SoftmaxActivation


class ActivationLayer:
    """The ActivationLayer class acts as a connector between a layer and an
    activation function. It ensures that the activation function is used 
    correctly during forward and backward propogation."""
    def __init__(self, activation):
        self.type = "Activation"
        layer_activation = activation
        self.activation = layer_activation.activation
        self.activation_prime = layer_activation.activation_partial_derivative

    def init_weights(self, input_size):
        pass

    def forward_propagation(self, input):
        """Applies the classes activation function to the provided input

        Args:
            input (np.array): output calculated forward up to this layer

        Returns:
            np.array(float): forward pass (output) up to this layer
        """
        self.input = input
        return self.activation(input)

    def backward_propagation(self, output_error, learning_rate):
        """Applies the classes activation function to the provided input

        Args:
            output_error (np.array): output_error calculated backwards to this layer

        Returns:
            np.array(float): backwards pass (output_error) up to this layer
        """
        return output_error * self.activation_prime(self.input)


class Sigmoid(ActivationLayer):
    """Layer that applies the Sigmoid activation function. Inheirits
    forward_propagation and backward_prop from ActivationLayer"""
    def __init__(self):
        super().__init__(SigmoidActivation)
        self.name = "Sigmoid"

class RELU(ActivationLayer):
    """Layer that applies the ReLU activation function. Inheirits
    forward_propagation and backward_prop from ActivationLayer"""
    def __init__(self):
        super().__init__(RELUActivation)
        self.name = "RELU"

class Softmax(ActivationLayer):
    """Layer that applies the Softmax activation function. Inheirits
    forward_propagation and backward_prop from ActivationLayer"""
    def __init__(self):
        super().__init__(SoftmaxActivation)
        self.name = "Softmax"


import numpy as np
from abc import ABC, abstractmethod
from gwu_nn.activation_layers import Sigmoid, RELU, Softmax

activation_functions = {'relu': RELU, 'sigmoid': Sigmoid, 'softmax': Softmax}


def apply_activation_forward(forward_pass):
    """Decorator that ensures that a layer's activation function is applied after the layer during forward
    propagation.
    """
    def wrapper(*args):
        output = forward_pass(args[0], args[1])
        if args[0].activation:
            return args[0].activation.forward_propagation(output)
        else:
            return output
    return wrapper


def apply_activation_backward(backward_pass):
    """Decorator that ensures that a layer's activation function's derivative is applied before the layer during
    backwards propagation.
    """
    def wrapper(*args):
        output_error = args[1]
        learning_rate = args[2]
        if args[0].activation:
            output_error = args[0].activation.backward_propagation(output_error, learning_rate)
        return backward_pass(args[0], output_error, learning_rate)
    return wrapper


class Layer():
    """The Layer layer is an abstract object used to define the template
    for other layer types to inherit"""

    def __init__(self, activation=None):
        """Because Layer is an abstract object, we don't provide any detailing
        on the initializtion"""
        self.type = "Layer"
        if activation:
            self.activation = activation_functions[activation]()
        else:
            self.activation = None

    @apply_activation_forward
    def forward_propagation(cls, input):
        """:noindex:"""
        pass

    @apply_activation_backward
    def backward_propogation(cls, output_error, learning_rate):
        """:noindex:"""
        pass


class Dense(Layer):
    """The Dense layer class creates a layer that is fully connected with the previous
    layer. This means that the number of weights will be MxN where M is number of
    nodes in the previous layer and N = number of nodes in the current layer.
    """

    def __init__(self, output_size, add_bias=False, activation=None, input_size=None):
        super().__init__(activation)
        self.type = None
        self.name = "Dense"
        self.input_size = input_size
        self.output_size = output_size
        self.add_bias = add_bias


    def init_weights(self, input_size):
        """Initialize the weights for the layer based on input and output size

        Args:
            input_size (numpy array): dimensions for the input array
        """
        if self.input_size is None:
            self.input_size = input_size

        self.weights = np.random.randn(input_size, self.output_size) / np.sqrt(input_size + self.output_size)
        if self.add_bias:
            self.bias = np.random.randn(1, self.output_size) / np.sqrt(input_size + self.output_size)


    @apply_activation_forward
    def forward_propagation(self, input):
        """Applies the forward propagation for a densely connected layer. This will compute the dot product between the
        input value (calculated during forward propagation) and the layer's weight tensor.

        Args:
            input (np.array): Input tensor calculated during forward propagation up to this layer.

        Returns:
            np.array(float): The dot product of the input and the layer's weight tensor."""
        self.input = input
        output = np.dot(input, self.weights)
        if self.add_bias:
            return output + self.bias
        else:
            return output

    @apply_activation_backward
    def backward_propagation(self, output_error, learning_rate):
        """Applies the backward propagation for a densely connected layer. This will calculate the output error
         (dot product of the output_error and the layer's weights) and will calculate the update gradient for the
         weights (dot product of the layer's input values and the output_error).

        Args:
            output_error (np.array): The gradient of the error up to this point in the network.

        Returns:
            np.array(float): The gradient of the error up to and including this layer."""
        input_error = np.dot(output_error, self.weights.T)
        weights_error = np.dot(self.input.T, output_error)

        self.weights -= learning_rate * weights_error
        if self.add_bias:
            self.bias -= learning_rate * output_error
        return input_error
